/*
 * MODIFICATION HISTORY
 *
 *	A000	vit@auriga.ru	May 6, 1998
 *		Porting to OpenServer.
 *		KEYS_FILE was moved to the include file oampkg.h. The part of
 *		code was excluded for OpenServer.
 */

#ident	"@(#)pkgactkey.c	15.1"

#include "oampkg.h"
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/keyctl.h>
#include <pfmt.h>

#define	PDLEN		4		/* Product id length. */
#define	ALLOC_SIZ	64
#define	MSG_NOARG	"uxpkgtools:796:- key argument is invalid\n"
#define	ERR_MEMORY	"uxpkgtools:6:memory allocation failure, errno=%d"

extern int getkeysid(char *, k_skey_t *);
extern int fgetkeysid(FILE *, char *, k_skey_t *);
extern int prid2list(char ***, char *, const char *);
extern int pkg_validate(k_skey_t *, char *, int);

/*
 * Get a valid serial number/activation key entry from a given file.
 * The first valid entry is returned if no product id is specified.
 * Otherwise the first matched entry from the current file offset is
 * returned.
 */
int
fgetkeysid(FILE *fp, char *prodid, k_skey_t *kp)
{
	int found;
	char *snum;
	register char *p;
	char buf[BUFSIZ];

	if (fp == NULL || prodid == NULL || kp == NULL) {
		return(0);
	}

	found = 0;
	kp->sernum[0] = '\0';
	kp->serkey[0] = '\0';
	while ((p = fgets(buf, BUFSIZ, fp)) != NULL) {
		if ((snum = strtok(p, " \t\n")) == NULL || *p == '#') {
			continue;
		}
		if ((p = strtok(NULL, " \t\n")) == NULL) {
			found = 0;
			break;
		}
		if (prodid[0] == '\0' ||
		    !strncmp(prodid, snum, strlen(prodid))) {
			found++;
			(void)strncpy((char *)kp->sernum,
					snum, STRLEN);
			(void)strncpy((char *)kp->serkey, p, STRLEN);
			break;
		}
	}
	return(found);
}

/*
 * This getkeysid() routine uses the /etc/config/licensekeys file
 * as the file containing the serial number/activation key pairs
 * of Unixware products.  It calls fgetkeysid() to get the entries.
 */
int
getkeysid(char *prodid, k_skey_t *kp)
{
	static FILE *fp = NULL;

	if (prodid == NULL) {
		/*
		 * This is a request to close the licensekeys file.
		 */
		if (fp) {
			(void)fclose(fp);
			fp = NULL;
		}
		return(1);
	}
	if (kp == NULL) {
		errno = EINVAL;
		pfmt(stderr, MM_ERROR, MSG_NOARG);
		return(-1);
	}
	if (fp == NULL) {
		if ((fp = fopen(KEYS_FILE, "r")) == NULL) {
			return(-1);
		}
	}
	if (prodid[0]) {
		if (fseek(fp, 0L, 0)) {
			return(-1);
		}
	}
	return(fgetkeysid(fp, prodid, kp) ? 0 : -1);
}

int
prid2list(char ***strtab, char *prid, const char *delim)
{
	int i;
	char **pstr;
	register char *pt;

	if (prid == NULL) {
		/*
		 * Free resources associated with strtab.
		 */
		pstr = *strtab;
		if (pstr != NULL && pstr[0] != NULL) {
			(void)free(pstr[0]);
			(void)free(pstr);
		} else if (pstr != NULL) {
			(void)free(pstr);
		}
		*strtab = NULL;
		return 0;
	}
	if ((pt = strdup(prid)) == NULL) {
		progerr(ERR_MEMORY, errno);
		return -1;
	}
	if ((pstr = (char **)calloc(ALLOC_SIZ, sizeof(char *))) == NULL) {
		progerr(ERR_MEMORY, errno);
		return -1;
	}
	i = 0;
	pt = strtok(pt, delim);
	while (pt != NULL) {
		pstr[i++] = pt;
		if ((i % ALLOC_SIZ) == 0) {
			pstr = (char **)realloc(pstr,
					(i + ALLOC_SIZ) * sizeof(char *));
			if (pstr == NULL) {
				progerr(ERR_MEMORY, errno);
				return -1;
			}
		}
		pt = strtok(NULL, delim);
	}
	pstr[i] = NULL;
	pstr = (char **)realloc(pstr, (i + 1) * sizeof(char *));
	*strtab = pstr;
	return pstr == NULL ? -1 : i;
}

#ifndef _SCO_DS							/* A000 */
int 
gen_key(k_skey_t *sp, int first, int last)
{
	/*
	 * The following sequence is in order of probability to try
	 * to minimize looping.
	 */
	char *chk_seq = "2510UXQCDMV346789";
	int n, ret;

	if ( first == last ) {
		return keyctl(K_VALIDATE, sp, 1);
	}

	last--;
	for (n = 0; chk_seq[n] != '\0'; n++) {
		sp->sernum[last] = chk_seq[n];
		if ((ret = gen_key( sp, first, last)) == 0) {
			return ret;
		}
	}
	return 1;
}

int
pkg_validate(k_skey_t *sp, char *prodname, int prefix)
{
	int ret = 1;
	int cnt;
	register int i, n;
	char **ptab;
	int pdlen;

	/*
	 * Convert product IDs to an array of strings.
	 */
	if ((cnt = prid2list(&ptab, prodname, "|")) <= 0) {
		return ret;
	}

	/*
	 * Here we assume that leading '0's indicate that the
	 * product id was not specified.
	 */
	for (pdlen = 0; pdlen < PDLEN; pdlen++) {
		if ( sp->sernum[pdlen] != '0' )
			break;
	}
	if (prefix && (pdlen != 0)) {
		/*
		 * The "serial number" may or may/not have the product
		 * identifier prefix, which must be supplied on the
		 * keyctl() call. If not supplied, try the values in
		 * the PRODUCTNAME package parameter.
		 */
		for (n = 0; n < cnt; n++) {
			if ( strlen(ptab[n]) > pdlen )
				continue;
			for (i = 0; i < pdlen, ptab[n][i]; i++) {
				sp->sernum[i] = ptab[n][i];
			}
			/*
			 * For unspecified product id locations,
			 * try the list of possible values, chk_seq.
			 */
			if ((ret = gen_key(sp, i, pdlen)) == 0)
				return ret;
			while (i-- >= 0) {
				sp->sernum[i] = '0';
			}
		}
		return ret;
	}
	/*
	 * Verify that the product ID part of the SERIALNUM
	 * exported environment variable matches at least one
	 * of the specified product IDs.
	 */
	for (n = 0; n < cnt; n++) {
		if (strncmp(ptab[n], (const char *)sp->sernum,
			strlen(ptab[n])) == 0) {
			break;
		}
	}
	/*
	 * Free memory resources associated with table.
	 */
	(void)prid2list(&ptab, NULL, NULL);
	return n >= cnt ? n : keyctl(K_VALIDATE, sp, 1);
}
#endif								/* A000 */
