#include <wait.h>
#include "sys/keyctl.h"
#include "oampkg.h"

#define BRAND "/etc/brand"

extern int prid2list(char ***, char *, const char *);

static int brand ( uchar_t * sernum, uchar_t * serkey )
{
  char cmd[PATH_MAX];
  int status;

  if ( sernum == 0 || serkey == 0 )
    return 0;

  sprintf( cmd, "%s -h %s %s >/dev/null 2>/dev/null", BRAND, sernum, serkey );
  status = system( cmd );

  if ( status == -1 )
    return 0;
  return WIFEXITED(status) && WEXITSTATUS(status) == 8;
}

int
pkg_validate(k_skey_t *sp, char *prodname, int prefix)
{
	int ret = 1;
	int cnt;
	register int i, n = 0;
	char **ptab;
	int pdlen;

	/*
	 * Convert product IDs to an array of strings.
	 */
	if ((cnt = prid2list(&ptab, prodname, "|")) <= 0) {
		return ret;
	}

	/*
	 * Verify that the product ID part of the SERIALNUM
	 * exported environment variable matches at least one
	 * of the specified product IDs.
	 */
	for (n = 0; n < cnt; n++) {
		if (strncmp(ptab[n], (const char *)sp->sernum,
			strlen(ptab[n])) == 0) {
			break;
		}
	}
	/*
	 * Free memory resources associated with table.
	 */
	(void)prid2list(&ptab, NULL, NULL);
	return n >= cnt ? n : !brand( sp->sernum, sp->serkey );
}
