/*LINTLIBRARY*/
#ident	"@(#)pkgtrans.c	15.1"
#ident "$Header: $"

#include "oampkg.h"
#include <stdio.h>
#include <varargs.h>
#include <limits.h>
#include <stdlib.h>
#include <ctype.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <wait.h>
#include <errno.h>
#include <dirent.h>
#include <signal.h>
#include <fcntl.h>
#include <pkginfo.h>
#include <pkgstrct.h>
#include <pkgtrans.h>
#include <pkgdev.h>

#include <pfmt.h>

extern int	errno;
extern void	progerr(),
		logerr(),
		rpterr(),
		ds_order();
extern int	getvol(),
		_getvol(),
		chdir(),
		access(),
		ds_init(),
                ds_gpkglist(),
		ds_findpkg(),
		ds_getpkg(),
		devtype(),
		ds_ginit(),
		ds_readbuf(),
		ds_close(),
		write(),
		dup(),
		close(),
		gpkgmap();

siginfo_t infop;

#define CMDSIZ	4096
#define LSIZE	512
#define TMPSIZ	64
#define PKGLEN	10

#define PKGINFO	"pkginfo"
#define PKGMAP	"pkgmap"
#define SETINFO	"setinfo"
#define INSTALL	"install"

#define ERR_LIST	"uxpkgtools:776:unable to complete pdatastream list"
#define ERR_REOPEN	"uxpkgtools:774:unable to reopen stdin"
#define MSG_CMDFAIL	"uxpkgtools:606:- process <%s> failed, exit code %d"
#define MSG_CMDBIG	"uxpkgtools:662:- command line too big - <%s>"
#define MSG_POPEN	"uxpkgtools:609:- popen of <%s> failed, errno=%d"
#define MSG_PCLOSE	"uxpkgtools:611:- pclose of <%s> failed, errno=%d"
#define MSG_BADDEV	"uxpkgtools:663:- invalid or unknown device <%s>"
#define MSG_GETVOL	"uxpkgtools:664:- unable to obtain package volume"
#define MSG_CHDIR	"uxpkgtools:666:- unable to change directory to <%s>"
#define MSG_NOPKGS	"uxpkgtools:673:- no packages were found in <%s>"
#define MSG_MKDIR	"uxpkgtools:674:- unable to make directory <%s>"

#define MSG_DSNM	"uxpkgtools:777:- no file or file is not a datastream"

static struct pkgdev srcdev, dstdev;
static char	*tmppath;
static char	dstinst[16];
static char 	*ids_name, *ods_name;
static char	*out_dev;
static int	ds_volcnt;
static int	ds_volno;
static void	(*func)();
static void	cleanup();
static int	pkgxfer();
int		pkgtrans();
extern int	ds_fd; /* open file descriptor for data stream */
static int	setinfo = 0;	/* flag for when setinfo exists */

char	**xpkg;	/* array of transferred packages */
int	nxpkg;
char	*pkgmap = NULL; /*variable needed for ESD(Electronic Software Distribution */

static	char *allpkg[] = {
	"all",
	NULL
};


/* will return 0, 1, 3, or 99 */
int
pkgls(device1, pkg, options)
char	*device1, **pkg;
int	options;
{
	char	*src, *dst;
	int	errflg, i, n;
	char	path[PATH_MAX];
	struct stat statbuf;


	/* transfer spool to appropriate device */
	if(devtype(device1, &srcdev)) {
		progerr(ERR_LIST);
		logerr(MSG_BADDEV, device1);
		return(1);
	}
	srcdev.rdonly++;


	/* check for datastream */
	ids_name = NULL;
	if(srcdev.bdevice) {
		if(n = _getvol(srcdev.bdevice, NULL, NULL, gettxt("uxpkgtools:137", "Insert %v into %p."), srcdev.norewind, 0)) {
			cleanup();
			if(n == 3)
				return(3);
			progerr(ERR_LIST);
			logerr(MSG_GETVOL);
			return(1);
		}
		if(ds_readbuf(srcdev.cdevice))
			ids_name = srcdev.cdevice;
	}

	if(srcdev.cdevice && !srcdev.bdevice) 
		ids_name = srcdev.cdevice;
	else if(srcdev.pathname) {
		ids_name = srcdev.pathname;
		if(strcmp(srcdev.pathname, "-") && access(srcdev.pathname, 0) == -1)
		{
			progerr(ERR_LIST);
			logerr(MSG_GETVOL);
			return(1);
		}
	}
		
        if (ids_name == NULL)
        {
		progerr(ERR_LIST);
		logerr(MSG_DSNM);
                return (1);
        }


	if(srcdev.cdevice && !srcdev.bdevice && 
    	    (n = _getvol(srcdev.cdevice, NULL, NULL, NULL, srcdev.norewind, 0)))
    	{
		cleanup();
		if(n == 3)
			return(3);
		progerr(ERR_LIST);
		logerr(MSG_GETVOL);
		return(1);
	}

	if(strcmp(ids_name, "-") == 0)
	{
		if(!pkg || pkg[0] == NULL) {
			ds_fd = fcntl(0, F_DUPFD, 1);
			if(freopen("/dev/tty", "r", stdin) == NULL) {
				progerr(ERR_LIST);
				logerr(ERR_REOPEN);
				cleanup();
				return(1);
			}
		}
	}
	if(ds_init(ids_name, pkg, srcdev.norewind)) {
		cleanup();
		return(1);
	}

	src = srcdev.dirname;
	dst = dstdev.dirname;


	if(src && chdir(src))
	{
		progerr(ERR_LIST);
		logerr(MSG_CHDIR, src);
		cleanup();
		return(1);
	}

        if (src == NULL)
            src = srcdev.pathname;


        errno=0;        
	if (*pkg)
	{
            for (nxpkg=0; pkg[nxpkg] && *pkg[nxpkg]; nxpkg++)
                    ;
	}
	else
	{
	    nxpkg = ds_gpkglist(src, &pkg);
	}
        if(nxpkg == 0 )
        {
	    if(errno) {
		progerr(ERR_LIST);
		logerr(MSG_NOPKGS, src);
		cleanup();
		return(1);
	    } else {
		cleanup();
		return(0);
	    }
	}   


	srcdev.cdevice = strdup(srcdev.pathname);
        

	errflg = 0;
	for(i=0; i<nxpkg; i++)
        {
            fprintf(stdout, "--- %s ---\n", pkg[i]);
            errflg = pkgxls(pkg[i], options);
            if (errflg) break;
	}

	cleanup();
	return(errflg);
}



static int
pkgxls(srcinst, options)
char	*srcinst;
int	options;
{
	struct pkginfo info;
	char	*pt, *src, *dst, *t_pkgmap;
	char	srcdir[PATH_MAX],
		cmd[CMDSIZ],
		buf[512];
	int	i, n, part, nparts;


	info.pkginst = NULL; /* required initialization */

	src = srcdev.dirname;
	dst = dstdev.dirname;



	/* list the datastream */
	/*
	 * transfer pkginfo & pkgmap first
	 */
	(void) sprintf(cmd, "%s -iutm", CPIOPROC);

	/*
	 * for each part of the package, use cpio() to
	 * unpack the archive into the destination directory
	 */
	nparts = ds_findpkg(srcdev.cdevice, srcinst);
	if(nparts < 0) {
		progerr(ERR_LIST);
		return(1);
	}
	for(part=1; part <= nparts;part++)
	{
		if(ds_lspkg(srcdev.cdevice, part, options))
		{
			progerr(ERR_LIST);
			return(1);
		}
		
	}
	return(0);
}


static void
cleanup()
{
	(void)ds_close(1);

}


